# Copyright (c) 2026 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.api.rpc.agentnotifiers import utils as notifier_utils
from neutron.tests import base as test_base


class _RPCNotifier:
    def method_1(self, *args, **kwargs):
        return 'method_1_return'


class TestRPCNotifierHandler(test_base.BaseTestCase):
    def test_call_existing_method(self):
        handler = notifier_utils.RPCNotifierHandler()
        handler.notifier_instance = _RPCNotifier()
        self.assertEqual('method_1_return',
                         handler.method_1(1, arg2=2))

    @mock.patch.object(notifier_utils, 'LOG')
    def test_call_non_existing_method(self, mock_log):
        handler = notifier_utils.RPCNotifierHandler()
        handler.notifier_instance = _RPCNotifier()
        self.assertIsNone(handler.method_2(1, arg2=2))
        mock_log.warning.assert_called_once_with(
            'Method %s is not implemented in the RPC notifier.', 'method_2')

    def test_call_no_rpc_instance(self):
        handler = notifier_utils.RPCNotifierHandler()
        self.assertIsNone(handler.method_any(1, arg2=2))



# Copyright 2025 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add unique constraint to the network segment range

Revision ID: b1bca967e19d
Revises: ad80a9f07c5c
Create Date: 2025-04-08 11:28:47.791807

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b1bca967e19d'
down_revision = 'd553edeb540f'

network_segment_range_network_type = sa.Enum(
    'vlan', 'vxlan', 'gre', 'geneve',
    name='network_segment_range_network_type')

TABLE_NAME = 'network_segment_ranges'
network_segment_range_table = sa.Table(
    TABLE_NAME, sa.MetaData(),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('default', sa.Boolean(), nullable=False),
    sa.Column('shared', sa.Boolean(), nullable=False),
    sa.Column('project_id', sa.String(length=255), nullable=True),
    sa.Column('network_type', network_segment_range_network_type,
              nullable=False),
    sa.Column('physical_network', sa.String(length=64), nullable=False,
              server_default=''),
    sa.Column('minimum', sa.Integer(), nullable=True),
    sa.Column('maximum', sa.Integer(), nullable=True),
    sa.Column('standard_attr_id', sa.BigInteger(), nullable=False),
)


def upgrade():
    unique_name = 'uniq_network_segment_ranges'
    unique_columns = ['default',
                      'network_type',
                      'physical_network',
                      'minimum',
                      'maximum',
                      ]

    inspect = sa.inspect(op.get_bind())
    unique_constraints = inspect.get_unique_constraints(TABLE_NAME)
    for unique_constraint in unique_constraints:
        if unique_constraint['name'] == unique_name:
            # The unique constraint already exists.
            return

    migrate_values()
    clear_duplicate_values()
    op.alter_column(TABLE_NAME, 'physical_network', nullable=False,
                    server_default='', existing_type=sa.String(64))
    op.create_unique_constraint(
        columns=unique_columns,
        constraint_name=unique_name,
        table_name=TABLE_NAME)


def clear_duplicate_values():
    session = sa.orm.Session(bind=op.get_bind())
    values = []
    for row in session.query(network_segment_range_table):
        id = row[0]
        item = {'default': row[2],
                'network_type': row[5],
                'physical_network': row[6],
                'minimum': row[7],
                'maximum': row[8]}
        if item not in values:
            values.append(item)
        else:
            session.execute(
                network_segment_range_table.delete().where(
                    network_segment_range_table.c.id == id))
    session.commit()


def migrate_values():
    session = sa.orm.Session(bind=op.get_bind())
    values = []
    for row in session.query(network_segment_range_table):
        values.append({'id': row[0],
                       'physical_network': row[6]})
    for value in values:
        physical_network = value['physical_network'] or ''
        session.execute(
             network_segment_range_table.update().values(
                 physical_network=physical_network
             ).where(network_segment_range_table.c.id == value['id']))
    session.commit()



# Copyright 2025 Samsung SDS. All Rights Reserved
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ipaddress

import netaddr

import neutron.db.models_v2 as mod
from neutron.db import network_ip_availability_db
from neutron_lib import constants
from neutron_lib.db import api as db_api

TOTAL_IPS_IN_SUBNET = 'total_ips_in_subnet'
TOTAL_IPS_IN_ALLOCATION_POOL = 'total_ips_in_allocation_pool'
USED_IPS_IN_SUBNET = 'used_ips_in_subnet'
USED_IPS_IN_ALLOCATION_POOL = 'used_ips_in_allocation_pool'
FIRST_IP = 'first_ip'
LAST_IP = 'last_ip'
IP_AVAILABILITY_DETAILS = 'ip_availability_details'


class IpAvailabilityDetailsDbMixin(
        network_ip_availability_db.IpAvailabilityMixin):
    """Mixin class to provide detailed IP availability information."""

    @classmethod
    @db_api.CONTEXT_READER
    def get_network_ip_availabilities(cls, context, filters=None):
        res = super().get_network_ip_availabilities(context, filters)
        subnet_total_ips = cls._generate_subnet_total_ips(context,
                                                          filters)
        subnet_used_ips = cls._generat_subnet_used_ips(context,
                                                       filters)
        return cls._add_details_to_result(res, subnet_total_ips,
                                          subnet_used_ips)

    @classmethod
    def _generate_subnet_total_ips(cls, context, filters):
        """Generates a dict whose key=subnet_id, value=total_ips_dict.
        total_ips_dict has key-value:
          - 'total_ips_in_subnet': subnet cidr
          - 'total_ips_in_allocation_pool': the sum of IPs in each allocation
            pools, 0 if there are no allocation pools in the subnet
        """
        subnet_cidr_alloc_query = cls._build_subnet_cidr_alloc_query(context,
                                                                     filters)

        subnet_totals_dict = {}
        for row in subnet_cidr_alloc_query:
            # Skip networks without subnets
            if not row.subnet_id:
                continue

            # CIDR data
            total_ips_in_subnet = netaddr.IPNetwork(
                row.cidr, version=row.ip_version).size
            if row.ip_version == constants.IP_VERSION_4:
                # Exclude network and broadcast addresses.
                total_ips_in_subnet -= 2

            # IPAllocationPool data
            total_ips_in_allocation_pool = 0
            if row.first_ip:
                pool_total = netaddr.IPRange(
                    netaddr.IPAddress(row.first_ip),
                    netaddr.IPAddress(row.last_ip)).size
                cur_total = (
                    subnet_totals_dict.get(row.subnet_id)
                    .get(TOTAL_IPS_IN_ALLOCATION_POOL)
                    if subnet_totals_dict.get(row.subnet_id) else 0
                )
                total_ips_in_allocation_pool = cur_total + pool_total

            subnet_totals_dict[row.subnet_id] = {
                TOTAL_IPS_IN_SUBNET: total_ips_in_subnet,
                TOTAL_IPS_IN_ALLOCATION_POOL: total_ips_in_allocation_pool
            }
        return subnet_totals_dict

    @classmethod
    @db_api.CONTEXT_READER
    def _build_subnet_cidr_alloc_query(cls, context, filters):
        query = context.session.query(
            *cls.common_columns,
            mod.IPAllocationPool.first_ip,
            mod.IPAllocationPool.last_ip,
        )
        query = query.outerjoin(
            mod.Subnet,
            mod.Network.id == mod.Subnet.network_id
        )
        query = query.outerjoin(
            mod.IPAllocationPool,
            mod.Subnet.id == mod.IPAllocationPool.subnet_id)
        return cls._adjust_query_for_filters(query, filters)

    @classmethod
    def _generat_subnet_used_ips(cls, context, filters):
        """Generates a dict whose key=subnet_id, value=used_ips_dict.
        used_ips_dict has key-value:
          - 'used_ips_in_subnet': the sum of used IPs in the subnet (does not
            consider allocation pools)
          - 'used_ips_in_allocation_pool': the sum of used IPs in each
            allocation pool, 0 in case of no allocation pools
        """
        allocation_pools_query = (
            cls._build_allocation_pools_query(context, filters))

        pools_dict = {}
        used_ips_dict = {}

        for row in allocation_pools_query:
            if pools_dict.get(row.subnet_id) is None:
                pools_dict[row.subnet_id] = []
            pools_dict[row.subnet_id].append({
                FIRST_IP: (int(ipaddress.ip_address(row.first_ip))
                           if row.first_ip else None),
                LAST_IP: (int(ipaddress.ip_address(row.last_ip))
                          if row.last_ip else None)
            })
            if used_ips_dict.get(row.subnet_id) is None:
                used_ips_dict[row.subnet_id] = {
                    USED_IPS_IN_SUBNET: 0,
                    USED_IPS_IN_ALLOCATION_POOL: 0
                }

        allocations_query = cls._build_allocations_query(context, filters)

        for row in allocations_query:
            if pools_dict.get(row.subnet_id) is None:
                continue

            used_ips_dict[row.subnet_id][USED_IPS_IN_SUBNET] += 1

            ip_address = int(ipaddress.ip_address(row.ip_address))
            for pool in pools_dict[row.subnet_id]:
                if pool[FIRST_IP] is None:
                    continue
                if pool[FIRST_IP] <= ip_address <= pool[LAST_IP]:
                    used_ips_dict[row.subnet_id][USED_IPS_IN_ALLOCATION_POOL]\
                        += 1

        return used_ips_dict

    @classmethod
    @db_api.CONTEXT_READER
    def _build_allocation_pools_query(cls, context, filters):
        query = context.session.query(
            *cls.common_columns,
            mod.IPAllocationPool.first_ip,
            mod.IPAllocationPool.last_ip
        )
        query = query.outerjoin(
            mod.Subnet,
            mod.Network.id == mod.Subnet.network_id
        )
        query = query.outerjoin(
            mod.IPAllocationPool,
            mod.Subnet.id == mod.IPAllocationPool.subnet_id)
        return cls._adjust_query_for_filters(query, filters)

    @classmethod
    @db_api.CONTEXT_READER
    def _build_allocations_query(cls, context, filters):
        query = context.session.query(
            *cls.common_columns,
            mod.IPAllocation.ip_address
        )
        query = query.outerjoin(
            mod.Subnet,
            mod.Network.id == mod.Subnet.network_id
        )
        query = query.join(
            mod.IPAllocation,
            mod.Subnet.id == mod.IPAllocation.subnet_id
        )
        return cls._adjust_query_for_filters(query, filters)

    @classmethod
    def _add_details_to_result(cls, res, subnet_total_ips, subnet_used_ips):
        for i, net in enumerate(res):
            net_details = {
                TOTAL_IPS_IN_SUBNET: 0,
                TOTAL_IPS_IN_ALLOCATION_POOL: 0,
                USED_IPS_IN_SUBNET: 0,
                USED_IPS_IN_ALLOCATION_POOL: 0
            }

            res_sub = net['subnet_ip_availability']
            for j, sub in enumerate(res_sub):
                sub_id = sub['subnet_id']
                sub_details = {
                    TOTAL_IPS_IN_SUBNET:
                        subnet_total_ips[sub_id][TOTAL_IPS_IN_SUBNET],
                    TOTAL_IPS_IN_ALLOCATION_POOL:
                        subnet_total_ips[sub_id][TOTAL_IPS_IN_ALLOCATION_POOL],
                    USED_IPS_IN_SUBNET:
                        subnet_used_ips[sub_id][USED_IPS_IN_SUBNET],
                    USED_IPS_IN_ALLOCATION_POOL:
                        subnet_used_ips[sub_id][USED_IPS_IN_ALLOCATION_POOL]
                }
                res_sub[j][IP_AVAILABILITY_DETAILS] = sub_details

                net_details[TOTAL_IPS_IN_SUBNET]\
                    += sub_details[TOTAL_IPS_IN_SUBNET]
                net_details[TOTAL_IPS_IN_ALLOCATION_POOL]\
                    += sub_details[TOTAL_IPS_IN_ALLOCATION_POOL]
                net_details[USED_IPS_IN_SUBNET]\
                    += sub_details[USED_IPS_IN_SUBNET]
                net_details[USED_IPS_IN_ALLOCATION_POOL]\
                    += sub_details[USED_IPS_IN_ALLOCATION_POOL]

            res[i][IP_AVAILABILITY_DETAILS] = net_details

        return res



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron._i18n import _

bgp_opts = [
    cfg.StrOpt('main_router_name',
               default='bgp-lr-main',
               help=_('Name of the main BGP router.')),
    cfg.IntOpt('main_router_vrf_id',
               default=42,
               help=_('VRF ID for the main BGP router.')),
    cfg.IntOpt('chassis_router_vrf_id',
               default=10,
               help=_('VRF ID for the chassis BGP router. This VRF ID is used '
                      'to learn default routes for the egress traffic out '
                      'from the chassis.')),
]


def get_main_router_name():
    return cfg.CONF.bgp.main_router_name


def get_main_router_vrf_id():
    return cfg.CONF.bgp.main_router_vrf_id


def get_chassis_router_vrf_id():
    return cfg.CONF.bgp.chassis_router_vrf_id


def register_opts(conf):
    conf.register_opts(bgp_opts, group='bgp')



# Copyright 2025 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.linux import utils as agent_utils
from neutron.agent.metadata import proxy_base
from neutron.common import metadata as common_metadata
from neutron.common.ovn import constants as ovn_const


LOG = logging.getLogger(__name__)


class MetadataProxyHandler(
        common_metadata.MetadataProxyHandlerBaseSocketServer):
    NETWORK_ID_HEADER = 'X-OVN-Network-ID'
    ROUTER_ID_HEADER = ''
    _conf = None
    _chassis = None
    _sb_idl = None

    def __init__(self, request, client_address, server):
        super().__init__(self._conf, has_cache=False, request=request,
                         client_address=client_address, server=server)

    @property
    def sb_idl(self):
        return self._sb_idl

    def get_port(self, remote_address, network_id=None, remote_mac=None,
                 router_id=None, skip_cache=False):
        ports = self.sb_idl.get_network_port_bindings_by_ip(network_id,
                                                            remote_address,
                                                            mac=remote_mac)
        num_ports = len(ports)
        if num_ports == 1:
            external_ids = ports[0].external_ids
            return (external_ids[ovn_const.OVN_DEVID_EXT_ID_KEY],
                    external_ids[ovn_const.OVN_PROJID_EXT_ID_KEY])
        if num_ports == 0:
            LOG.warning("No port found in network %s with IP address %s",
                        network_id, remote_address)
        elif num_ports > 1:
            port_uuids = ', '.join([str(port.uuid) for port in ports])
            LOG.error("More than one port found in network %s with IP address "
                      "%s. Please run the neutron-ovn-db-sync-util script as "
                      "there seems to be inconsistent data between Neutron "
                      "and OVN databases. OVN Port uuids: %s", network_id,
                      remote_address, port_uuids)
        return None, None


class UnixDomainMetadataProxy(proxy_base.UnixDomainMetadataProxyBase):
    def __init__(self, conf, chassis, sb_idl=None):
        super().__init__(conf)
        self.chassis = chassis
        self.sb_idl = sb_idl
        agent_utils.ensure_directory_exists_without_file(
            cfg.CONF.metadata_proxy_socket)
        self._server = None

    def run(self):
        # Set the default metadata_workers if not yet set in the config file
        md_workers = self.conf.metadata_workers
        md_workers = 0 if md_workers is None else md_workers

        file_socket = cfg.CONF.metadata_proxy_socket
        self._server = proxy_base.MetadataProxyServer(
            md_workers, file_socket, MetadataProxyHandler)

        MetadataProxyHandler._conf = self.conf
        MetadataProxyHandler._chassis = self.chassis
        MetadataProxyHandler._sb_idl = self.sb_idl

    def wait(self):
        self._server.serve_forever()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron import pecan_wsgi
from neutron.tests import base


@mock.patch('neutron.pecan_wsgi.app.versions_factory')
class TestVersionsFactory(base.BaseTestCase):

    def test_versions_factory(self, pecan_mock):
        pecan_wsgi.app.versions_factory({})
        pecan_mock.assert_called_once_with({})



# Copyright 2022 Canonical
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.api.definitions import l3
from neutron_lib import constants as const
from neutron_lib import context as ncontext
from neutron_lib.services.logapi import constants as log_const
from neutron_lib.services.trunk import constants as trunk_const
from oslo_config import cfg

from neutron.common.ovn import constants
from neutron.conf.plugins.ml2 import config as ml2_conf
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.plugins.ml2 import db as ml2_db
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovn_client
from neutron.tests import base
from neutron.tests.unit import fake_resources as fakes
from neutron.tests.unit.services.logapi.drivers.ovn \
    import test_driver as test_log_driver

from tenacity import wait_none


class Test_has_separate_snat_per_subnet(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        ovn_conf.register_opts()

    def test_snat_on_nested_off(self):
        fake_router = {
            'id': 'fake-id',
            'enable_snat': True,
            l3.EXTERNAL_GW_INFO: mock.Mock(),  # irrelevant value
        }
        # ovn_router_indirect_snat default is False
        self.assertTrue(ovn_client._has_separate_snat_per_subnet(fake_router))

    def test_snat_off_nested_off(self):
        fake_router = {
            'id': 'fake-id',
            'enable_snat': False,
            l3.EXTERNAL_GW_INFO: mock.Mock(),  # irrelevant value
        }
        # ovn_router_indirect_snat default is False
        self.assertFalse(ovn_client._has_separate_snat_per_subnet(fake_router))

    def test_snat_on_nested_on(self):
        fake_router = {
            'id': 'fake-id',
            'enable_snat': True,
            l3.EXTERNAL_GW_INFO: mock.Mock(),  # irrelevant value
        }
        cfg.CONF.set_override('ovn_router_indirect_snat', True, 'ovn')
        self.assertFalse(ovn_client._has_separate_snat_per_subnet(fake_router))

    def test_snat_off_nested_on(self):
        fake_router = {
            'id': 'fake-id',
            'enable_snat': False,
            l3.EXTERNAL_GW_INFO: mock.Mock(),  # irrelevant value
        }
        cfg.CONF.set_override('ovn_router_indirect_snat', True, 'ovn')
        self.assertFalse(ovn_client._has_separate_snat_per_subnet(fake_router))


class TestOVNClientBase(base.BaseTestCase):

    def setUp(self):
        ml2_conf.register_ml2_plugin_opts()
        ovn_conf.register_opts()
        super().setUp()
        self.nb_idl = mock.MagicMock()
        self.sb_idl = mock.MagicMock()
        self.ovn_client = ovn_client.OVNClient(self.nb_idl, self.sb_idl)


class TestOVNClient(TestOVNClientBase):

    def setUp(self):
        super().setUp()
        self.get_plugin = mock.patch(
            'neutron_lib.plugins.directory.get_plugin').start()
        self.get_pb_bsah = mock.patch(
            'neutron_lib.plugins.utils.'
            'get_port_binding_by_status_and_host').start()

        # Disable tenacity wait for UT
        self.ovn_client._wait_for_active_port_bindings_host.retry.wait = (
            wait_none())

    def test__add_router_ext_gw_default_route(self):
        plugin = mock.MagicMock()
        self.get_plugin.return_value = plugin
        subnet = {
            'id': 'fake-subnet-id',
            'gateway_ip': '10.42.0.1',
            'ip_version': const.IP_VERSION_4,
        }
        plugin.get_subnet.return_value = subnet
        plugin.get_subnets_by_network.return_value = [subnet]
        router = {
            'id': 'fake-router-id',
            'gw_port_id': 'fake-port-id',
            'enable_snat': True,
        }
        txn = mock.MagicMock()
        self.ovn_client._get_router_gw_ports = mock.MagicMock()
        gw_port = fakes.FakePort().create_one_port(
            attrs={
                'id': router['gw_port_id'],
                'fixed_ips': [{
                    'subnet_id': subnet.get('id'),
                    'ip_address': '10.42.0.42'}]
            })
        self.ovn_client._get_router_gw_ports.return_value = [gw_port]
        self.assertEqual(
            [self.get_plugin().get_port()],
            self.ovn_client._add_router_ext_gw(mock.Mock(), router, txn))
        self.nb_idl.add_static_route.assert_called_once_with(
            'neutron-' + router['id'],
            ip_prefix=const.IPv4_ANY,
            nexthop='10.42.0.1',
            maintain_bfd=False,
            external_ids={
                'neutron:is_ext_gw': 'true',
                'neutron:subnet_id': subnet['id'],
                constants.OVN_LRSR_EXT_ID_KEY: 'true'})

    def test__add_router_ext_gw_default_route_ecmp(self):
        plugin = mock.MagicMock()
        self.get_plugin.return_value = plugin
        subnet1 = {
            'id': 'fake-subnet-id-1',
            'gateway_ip': '10.42.0.1',
            'ip_version': const.IP_VERSION_4,
        }
        subnet2 = {
            'id': 'fake-subnet-id-2',
            'gateway_ip': '10.42.42.1',
            'ip_version': const.IP_VERSION_4,
        }
        plugin.get_subnet.side_effect = [subnet1, subnet2, subnet1, subnet2]
        plugin.get_subnets_by_network.return_value = [subnet1, subnet2]
        router = {
            'id': 'fake-router-id',
            'gw_port_id': 'fake-port-id',
            'enable_default_route_ecmp': True,
            'enable_snat': True,
        }
        txn = mock.MagicMock()
        self.ovn_client._get_router_gw_ports = mock.MagicMock()
        gw_port1 = fakes.FakePort().create_one_port(
            attrs={
                'id': router['gw_port_id'],
                'fixed_ips': [{
                    'subnet_id': subnet1.get('id'),
                    'ip_address': '10.42.0.42'}]
            })
        gw_port2 = fakes.FakePort().create_one_port(
            attrs={
                'id': 'gw-port-id-2',
                'fixed_ips': [{
                    'subnet_id': subnet2.get('id'),
                    'ip_address': '10.42.42.42'}]
            })
        self.ovn_client._get_router_gw_ports.return_value = [
            gw_port1, gw_port2]
        self.assertEqual(
            [self.get_plugin().get_port(), self.get_plugin().get_port()],
            self.ovn_client._add_router_ext_gw(mock.Mock(), router, txn))
        self.nb_idl.add_static_route.assert_has_calls([
            mock.call('neutron-' + router['id'],
                      ip_prefix=const.IPv4_ANY,
                      nexthop=subnet1['gateway_ip'],
                      maintain_bfd=False,
                      external_ids={
                         'neutron:is_ext_gw': 'true',
                         'neutron:subnet_id': subnet1['id'],
                         constants.OVN_LRSR_EXT_ID_KEY: 'true'},
                      ),
            mock.call('neutron-' + router['id'],
                      ip_prefix=const.IPv4_ANY,
                      nexthop=subnet2['gateway_ip'],
                      maintain_bfd=False,
                      external_ids={
                         'neutron:is_ext_gw': 'true',
                         'neutron:subnet_id': subnet2['id'],
                         constants.OVN_LRSR_EXT_ID_KEY: 'true'},
                      ),
        ])

    def test__add_router_ext_gw_no_default_route(self):
        plugin = mock.MagicMock()
        self.get_plugin.return_value = plugin
        subnet = {
            'id': 'fake-subnet-id',
            'gateway_ip': None,
            'ip_version': const.IP_VERSION_4
        }
        plugin.get_subnet.return_value = subnet
        plugin.get_subnets_by_network.return_value = [subnet]
        router = {
            'id': 'fake-router-id',
            l3.EXTERNAL_GW_INFO: {
                'external_fixed_ips': [{
                        'subnet_id': subnet.get('id'),
                        'ip_address': '10.42.0.42'}],
            },
            'gw_port_id': 'fake-port-id',
            'enable_snat': True,
        }
        txn = mock.MagicMock()
        self.ovn_client._get_router_gw_ports = mock.MagicMock()
        gw_port = fakes.FakePort().create_one_port(
            attrs={
                'id': router['gw_port_id'],
                'fixed_ips': [{
                    'subnet_id': subnet.get('id'),
                    'ip_address': '10.42.0.42'}]
            })
        self.ovn_client._get_router_gw_ports.return_value = [gw_port]
        self.assertEqual(
            [self.get_plugin().get_port()],
            self.ovn_client._add_router_ext_gw(mock.Mock(), router, txn))
        self.nb_idl.add_static_route.assert_not_called()

    def test_checkout_ip_list(self):
        addresses = ["192.168.2.2/32", "2001:db8::/32"]
        add_map = self.ovn_client._checkout_ip_list(addresses)
        self.assertEqual(["192.168.2.2/32"], add_map[const.IP_VERSION_4])
        self.assertEqual(["2001:db8::/32"], add_map[const.IP_VERSION_6])

    def test_update_lsp_host_info_up(self):
        context = mock.MagicMock()
        host_id = 'fake-binding-host-id'
        port_id = 'fake-port-id'
        port_binding = mock.Mock(host=host_id)
        db_port = mock.Mock(id=port_id, port_bindings=[port_binding])
        self.get_pb_bsah.return_value = port_binding

        self.ovn_client.update_lsp_host_info(context, db_port)

        self.nb_idl.db_set.assert_called_once_with(
            'Logical_Switch_Port', port_id,
            ('external_ids', {constants.OVN_HOST_ID_EXT_ID_KEY: host_id}))

    def test_update_lsp_host_info_up_retry(self):
        context = mock.MagicMock()
        host_id = 'fake-binding-host-id'
        port_id = 'fake-port-id'
        port_binding = mock.Mock(host=host_id)
        port_binding_no_host = mock.Mock(host="")
        db_port_no_host = mock.Mock(
            id=port_id, port_bindings=[port_binding_no_host])
        self.get_pb_bsah.return_value = None

        with mock.patch.object(
                self.ovn_client,
                '_wait_for_active_port_bindings_host') as mock_wait:
            mock_wait.return_value = port_binding
            self.ovn_client.update_lsp_host_info(context, db_port_no_host)

            # Assert _wait_for_port_bindings_host was called
            mock_wait.assert_called_once_with(context, port_id)

        # Assert host_id was set
        self.nb_idl.db_set.assert_called_once_with(
            'Logical_Switch_Port', port_id,
            ('external_ids', {constants.OVN_HOST_ID_EXT_ID_KEY: host_id}))

    def test_update_lsp_host_info_up_retry_fail(self):
        context = mock.MagicMock()
        port_id = 'fake-port-id'
        db_port_no_host = mock.Mock(
            id=port_id, port_bindings=[mock.Mock(host="")])
        self.get_pb_bsah.return_value = None

        with mock.patch.object(
                self.ovn_client,
                '_wait_for_active_port_bindings_host') as mock_wait:
            mock_wait.side_effect = RuntimeError("boom")
            self.ovn_client.update_lsp_host_info(context, db_port_no_host)

            # Assert _wait_for_port_bindings_host was called
            mock_wait.assert_called_once_with(context, port_id)

        # Assert host_id was NOT set
        self.assertFalse(self.nb_idl.db_set.called)

    def test_update_lsp_host_info_down(self):
        context = mock.MagicMock()
        port_id = 'fake-port-id'
        db_port = mock.Mock(id=port_id)
        self.nb_idl.lsp_get_up.return_value.execute.return_value = False

        self.ovn_client.update_lsp_host_info(context, db_port, up=False)

        self.nb_idl.db_remove.assert_called_once_with(
            'Logical_Switch_Port', port_id, 'external_ids',
            constants.OVN_HOST_ID_EXT_ID_KEY, if_exists=True)

    def test_update_lsp_host_info_trunk_subport(self):
        context = mock.MagicMock()
        db_port = mock.Mock(id='fake-port-id',
                            device_owner=trunk_const.TRUNK_SUBPORT_OWNER)

        self.ovn_client.update_lsp_host_info(context, db_port)
        self.nb_idl.db_remove.assert_not_called()
        self.nb_idl.db_set.assert_not_called()

    @mock.patch.object(ml2_db, 'get_port')
    def test__wait_for_active_port_bindings_host(self, mock_get_port):
        context = mock.MagicMock()
        host_id = 'fake-binding-host-id'
        port_id = 'fake-port-id'
        port_binding = mock.Mock(host=host_id)
        port_binding_no_host = mock.Mock(host="")
        db_port_no_host = mock.Mock(
            id=port_id, port_bindings=[port_binding_no_host])
        db_port = mock.Mock(
            id=port_id, port_bindings=[port_binding])
        # no active binding, no binding with host, binding with host
        self.get_pb_bsah.side_effect = (None, port_binding_no_host,
                                        port_binding)

        mock_get_port.side_effect = (db_port_no_host, db_port_no_host, db_port)

        ret = self.ovn_client._wait_for_active_port_bindings_host(
            context, port_id)

        self.assertEqual(ret, port_binding)

        expected_calls = [mock.call(context, port_id),
                          mock.call(context, port_id)]
        mock_get_port.assert_has_calls(expected_calls)

    @mock.patch.object(ml2_db, 'get_port')
    def test__wait_for_active_port_bindings_host_fail(self, mock_get_port):
        context = mock.MagicMock()
        port_id = 'fake-port-id'
        db_port_no_pb = mock.Mock(id=port_id, port_bindings=[])
        db_port_no_host = mock.Mock(
            id=port_id, port_bindings=[mock.Mock(host="")])
        self.get_pb_bsah.return_value = None

        mock_get_port.side_effect = (
            db_port_no_pb, db_port_no_host, db_port_no_host)

        self.assertRaises(
            RuntimeError, self.ovn_client._wait_for_active_port_bindings_host,
            context, port_id)

        expected_calls = [mock.call(context, port_id),
                          mock.call(context, port_id),
                          mock.call(context, port_id)]
        mock_get_port.assert_has_calls(expected_calls)

    def test__get_snat_cidrs_for_external_router_nested_snat_off(self):
        ctx = ncontext.Context()
        per_subnet_cidrs = ['10.0.0.0/24', '20.0.0.0/24']
        with mock.patch.object(
                self.ovn_client, '_get_v4_network_of_all_router_ports',
                return_value=per_subnet_cidrs):
            cidrs = self.ovn_client._get_snat_cidrs_for_external_router(
                ctx, 'fake-id')
        self.assertEqual(per_subnet_cidrs, cidrs)

    def test__get_snat_cidrs_for_external_router_nested_snat_on(self):
        ctx = ncontext.Context()
        cfg.CONF.set_override('ovn_router_indirect_snat', True, 'ovn')
        per_subnet_cidrs = ['10.0.0.0/24', '20.0.0.0/24']
        with mock.patch.object(
                self.ovn_client, '_get_v4_network_of_all_router_ports',
                return_value=per_subnet_cidrs):
            cidrs = self.ovn_client._get_snat_cidrs_for_external_router(
                ctx, 'fake-id')
        self.assertEqual([const.IPv4_ANY], cidrs)


class TestOVNClientFairMeter(TestOVNClientBase,
                             test_log_driver.TestOVNDriverBase):

    def test_create_ovn_fair_meter(self):
        mock_find_rows = mock.Mock()
        mock_find_rows.execute.return_value = None
        self.nb_idl.db_find_rows.return_value = mock_find_rows
        self.ovn_client.create_ovn_fair_meter(self._log_driver.meter_name)
        self.assertFalse(self.nb_idl.meter_del.called)
        self.assertTrue(self.nb_idl.meter_add.called)
        self.nb_idl.meter_add.assert_any_call(
            name=self._log_driver.meter_name + "_stateless",
            unit="pktps",
            rate=int(self.fake_cfg_network_log.rate_limit / 2),
            fair=True,
            burst_size=int(self.fake_cfg_network_log.burst_limit / 2),
            may_exist=False,
            external_ids={constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                          log_const.LOGGING_PLUGIN})
        self.nb_idl.meter_add.assert_any_call(
            name=self._log_driver.meter_name,
            unit="pktps",
            rate=self.fake_cfg_network_log.rate_limit,
            fair=True,
            burst_size=self.fake_cfg_network_log.burst_limit,
            may_exist=False,
            external_ids={constants.OVN_DEVICE_OWNER_EXT_ID_KEY:
                          log_const.LOGGING_PLUGIN})

    def test_create_ovn_fair_meter_unchanged(self):
        mock_find_rows = mock.Mock()
        fake_meter1 = [self._fake_meter()]
        fake_meter2 = [self._fake_meter(
            name=self._log_driver.meter_name + "_stateless",
            bands=[mock.Mock(uuid='tb_stateless')])]
        mock_find_rows.execute.side_effect = [fake_meter1, fake_meter1,
                                              fake_meter2, fake_meter2]
        self.nb_idl.db_find_rows.return_value = mock_find_rows
        self.nb_idl.lookup.side_effect = lambda table, key, default: (
            self._fake_meter_band() if key == "test_band" else
            self._fake_meter_band_stateless() if key == "tb_stateless" else
            default)
        self.ovn_client.create_ovn_fair_meter(self._log_driver.meter_name)
        self.assertFalse(self.nb_idl.meter_del.called)
        self.assertFalse(self.nb_idl.meter_add.called)

    def test_create_ovn_fair_meter_changed(self):
        mock_find_rows = mock.Mock()
        mock_find_rows.execute.return_value = [self._fake_meter(fair=[False])]
        self.nb_idl.db_find_rows.return_value = mock_find_rows
        self.nb_idl.lookup.return_value = self._fake_meter_band()
        self.ovn_client.create_ovn_fair_meter(self._log_driver.meter_name)
        self.assertTrue(self.nb_idl.meter_del.called)
        self.assertTrue(self.nb_idl.meter_add.called)

    def test_create_ovn_fair_meter_band_changed(self):
        mock_find_rows = mock.Mock()
        mock_find_rows.execute.return_value = [self._fake_meter()]
        self.nb_idl.db_find_rows.return_value = mock_find_rows
        self.nb_idl.lookup.return_value = self._fake_meter_band(rate=666)
        self.ovn_client.create_ovn_fair_meter(self._log_driver.meter_name)
        self.assertTrue(self.nb_idl.meter_del.called)
        self.assertTrue(self.nb_idl.meter_add.called)

    def test_create_ovn_fair_meter_band_missing(self):
        mock_find_rows = mock.Mock()
        mock_find_rows.execute.return_value = [self._fake_meter()]
        self.nb_idl.db_find_rows.return_value = mock_find_rows
        self.nb_idl.lookup.side_effect = None
        self.ovn_client.create_ovn_fair_meter(self._log_driver.meter_name)
        self.assertTrue(self.nb_idl.meter_del.called)
        self.assertTrue(self.nb_idl.meter_add.called)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.schema.ovn_southbound import impl_idl as sb_impl_idl

from neutron.common import utils as common_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.conf.services import bgp as bgp_config
from neutron.services.bgp import constants
from neutron.services.bgp import helpers
from neutron.services.bgp import ovn as bgp_ovn
from neutron.services.bgp import reconciler
from neutron.tests.functional import base
from neutron.tests.functional.services import bgp as bgp_tests


class TestBGPReconciler(base.TestOVNFunctionalBase):
    def setUp(self):
        ovn_conf.register_opts()
        bgp_config.register_opts(cfg.CONF)
        sb_impl_idl.OvnSbApiIdlImpl._ovsdb_connection = None
        bgp_ovn.BgpOvnSbIdl._ovsdb_connection = None
        bgp_ovn.BgpOvnNbIdl._ovsdb_connection = None
        super().setUp()
        self.reconciler = reconciler.BGPTopologyReconciler()

        if not bgp_tests.is_policy_output_port_column_supported(
                self.reconciler.nb_api.idl):
            raise self.skipException("Policy output port column not supported")

        self.nb_api = self.reconciler.nb_api
        self.sb_api = self.reconciler.sb_api

        self.test_sb_idl = self._create_additional_sb_idl()
        self.chassis_bgp_networks = ('bgp-net-1', 'bgp-net-2')

    @staticmethod
    def _create_additional_sb_idl():
        """Create a secondary SB IDL for testing.

        The sb_api instance does not register the Encap table and can't be used
        to create a chassis. This secondary SB IDL is there to control the
        chassis used for testing.
        """
        tables = ['Chassis', 'Chassis_Private', 'Encap']
        connection_string = ovn_conf.get_ovn_sb_connection()
        idl = connection.OvsdbIdl.from_server(
            connection_string, 'OVN_Southbound', helper_tables=tables)
        conn = connection.Connection(idl, timeout=10)
        return sb_impl_idl.OvnSbApiIdlImpl(conn)

    def _create_chassis(self, chassis_name, ip, bgp_bridges=None):
        external_ids = {}
        chassis = self.test_sb_idl.chassis_add(
            chassis_name,
            ['geneve'],
            ip,
            hostname=chassis_name,
        ).execute(check_error=True)

        if bgp_bridges:
            external_ids[
                constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY] = ','.join(
                    bgp_bridges)
        self.sb_api.db_create(
            'Chassis_Private', name=chassis_name,
            chassis=chassis.uuid,
            external_ids=external_ids
        ).execute(check_error=True)

        return self.sb_api.db_list_rows(
            'Chassis_Private', [chassis_name]).execute(check_error=True)[0]

    def _get_all_chassis_private(self):
        return self.sb_api.db_list_rows('Chassis_Private').execute(
            check_error=True)

    def validate_topology(self, chassis):
        self._validate_main_router()
        for ch in chassis:
            self._validate_chassis(ch)

    def _validate_main_router(self):
        main_router_name = bgp_config.get_main_router_name()
        main_router = self.nb_api.lr_get(
            main_router_name).execute(check_error=True)
        all_chassis_names = {
            chassis.name for chassis in self._get_all_chassis_private()}
        for chassis_name in list(all_chassis_names):
            chassis_router_name = helpers.get_chassis_router_name(chassis_name)
            lrp_to_chassis_router_name = helpers.get_lrp_name(
                main_router_name, chassis_router_name)
            lrp = self.nb_api.lrp_get(
                lrp_to_chassis_router_name).execute(check_error=True)
            self.assertIn(lrp, main_router.ports)

            try:
                ha_chassis = lrp.ha_chassis_group[0].ha_chassis[0]
            except IndexError:
                self.fail(
                    f"LRP {lrp.name} on the main BGP router has no chassis "
                    f"binding: {lrp.ha_chassis_group}")

            self.assertEqual(ha_chassis.chassis_name, chassis_name)
            try:
                all_chassis_names.remove(chassis_name)
            except KeyError:
                self.fail(f"Chassis {chassis_name} not found for LRP "
                          f"{lrp_to_chassis_router_name} on the main BGP "
                          f"router")
        if all_chassis_names:
            self.fail(f"There are some chassis remaining that do not have an "
                      f"LRP from the main BGP router bound to it: "
                      f"{all_chassis_names}")

    def _validate_chassis(self, chassis):
        router_name = helpers.get_chassis_router_name(chassis.name)
        router = self.nb_api.lr_get(router_name).execute(check_error=True)

        self.assertEqual(chassis.name, router.options.get('chassis'))
        # each router has two connections out and one to the main BGP router
        self.assertEqual(3, len(router.ports))

        # each router should be connected to two switches with name format
        # bgp-ls-<chassis_name>-<network_name> where network names are from
        # the bgp_peer_connections: bgp-net-1 and bgp-net-2
        for network_name in self.chassis_bgp_networks:
            switch_name = helpers.get_chassis_peer_switch_name(
                chassis.name, network_name)
            ls = self.nb_api.ls_get(switch_name).execute(check_error=True)

            # each switch should have a localnet port with the network name set
            # and one port plugged to the router
            self.assertEqual(2, len(ls.ports))
            for lsp in ls.ports:
                if lsp.type == 'localnet':
                    self.assertEqual(
                        network_name, lsp.options.get('network_name'))
                elif lsp.type == 'router':
                    lrp = self.nb_api.lrp_get(
                        lsp.options.get('router-port')).execute(
                            check_error=True)
                    self.assertIn(lrp, router.ports)
                    ext_ids = lrp.external_ids
                    self.assertEqual(
                        network_name, ext_ids[
                            constants.LRP_NETWORK_NAME_EXT_ID_KEY])

    def test_full_sync(self):
        for i in range(0, 6):
            chassis_name = f'chassis{i}'
            self._create_chassis(
                chassis_name, f'192.168.1.10{i}',
                bgp_bridges=self.chassis_bgp_networks)
        self.reconciler.full_sync()

        self.validate_topology(self._get_all_chassis_private())

    def test_setting_chassis_bgp_bridges_configures_lrps(self):
        chassis = self._create_chassis(
            'chassis', '192.168.1.100')

        self.reconciler.full_sync()

        # we have an environment with one chassis without bgp peer connections
        # set
        chassis = self._get_all_chassis_private()[0]

        # There should be no switches in the environment
        switches = self.nb_api.db_list_rows('Logical_Switch').execute(
            check_error=True)
        self.assertEqual(0, len(switches))

        external_ids = {constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: ','.join(
            self.chassis_bgp_networks)}

        self.test_sb_idl.db_set(
            'Chassis_Private', 'chassis', external_ids=external_ids).execute(
                check_error=True)

        def switches_created():
            return len(self.nb_api.db_list_rows('Logical_Switch').execute(
                check_error=True)) == 2

        common_utils.wait_until_true(
            switches_created,
            timeout=10,
            exception=Exception(
                "Peer switches were not created")
        )

        self.validate_topology([chassis])



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import uuid

from neutron_lib.db import api as db_api
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from neutron.db.models import ovn as ovn_models

CONF = cfg.CONF
LOG = log.getLogger(__name__)

OVN_HASHRING_UUID_NAMESPACE = uuid.UUID('551ea055-1982-4295-8bc7-207a0d3c9231')


def get_node_uuid(
        group_name: str,
        host: str,
        worker_id: int) -> str:
    node_str = '{}{}{}'.format(group_name, host, str(worker_id))
    return uuid.uuid5(OVN_HASHRING_UUID_NAMESPACE, node_str).hex


# NOTE(ralonsoh): this was migrated from networking-ovn to neutron and should
#                 be refactored to be integrated in a OVO.
@db_api.retry_if_session_inactive()
def add_node(context, group_name, node_uuid, created_at=None):
    with db_api.CONTEXT_WRITER.using(context):
        kwargs = {'node_uuid': node_uuid,
                  'hostname': CONF.host,
                  'group_name': group_name}
        if created_at:
            kwargs['created_at'] = created_at
        context.session.add(ovn_models.OVNHashRing(**kwargs))
    LOG.info('Node %s from host "%s" and group "%s" added to the Hash Ring',
             node_uuid, CONF.host, group_name)
    return node_uuid


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_READER
def get_nodes(context, group_name, created_at=None):
    query = context.session.query(ovn_models.OVNHashRing).filter(
        ovn_models.OVNHashRing.group_name == group_name)
    if created_at:
        query = query.filter(
            ovn_models.OVNHashRing.created_at == created_at)
    return query.all()


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_READER
def get_node(context, group_name, node_uuid):
    return context.session.query(ovn_models.OVNHashRing).filter(
        ovn_models.OVNHashRing.group_name == group_name,
        ovn_models.OVNHashRing.node_uuid == node_uuid).one()


@db_api.retry_if_session_inactive()
def remove_nodes_from_host(context, group_name, created_at=None):
    with (db_api.CONTEXT_WRITER.using(context)):
        query = context.session.query(ovn_models.OVNHashRing).filter(
            ovn_models.OVNHashRing.hostname == CONF.host,
            ovn_models.OVNHashRing.group_name == group_name)
        if created_at:
            query = query.filter(
                ovn_models.OVNHashRing.created_at != created_at)
        query.delete()
    msg = ('Nodes from host "%s" and group "%s" removed from the Hash Ring' %
           (CONF.host, group_name))
    if created_at:
        msg += ' created at %s' % str(created_at)
    LOG.info(msg)


@db_api.retry_if_session_inactive()
def remove_node_by_uuid(context, node_uuid):
    with db_api.CONTEXT_WRITER.using(context):
        context.session.query(ovn_models.OVNHashRing).filter(
            ovn_models.OVNHashRing.node_uuid == node_uuid).delete()
    LOG.info('Node "%s" removed from the Hash Ring', node_uuid)


@db_api.retry_if_session_inactive()
def cleanup_old_nodes(context, days):
    age = timeutils.utcnow() - datetime.timedelta(days=days)
    with db_api.CONTEXT_WRITER.using(context):
        context.session.query(ovn_models.OVNHashRing).filter(
            ovn_models.OVNHashRing.updated_at < age).delete()
    LOG.info('Cleaned up Hash Ring nodes older than %d days', days)


@db_api.CONTEXT_WRITER
def touch_node(context, node_uuid):
    # NOTE(ralonsoh): there are several mechanisms to update the node OVN hash
    # ring register. This method does not retry the DB operation in case of
    # failure but relies on the success of later calls. That will prevent from
    # blocking the DB needlessly.
    context.session.query(ovn_models.OVNHashRing).filter(
        ovn_models.OVNHashRing.node_uuid == node_uuid).update(
        {'updated_at': timeutils.utcnow()})


def _get_nodes_query(context, interval, group_name, offline=False,
                     from_host=False):
    limit = timeutils.utcnow() - datetime.timedelta(seconds=interval)
    query = context.session.query(ovn_models.OVNHashRing).filter(
        ovn_models.OVNHashRing.group_name == group_name)

    if offline:
        query = query.filter(ovn_models.OVNHashRing.updated_at < limit)
    else:
        query = query.filter(ovn_models.OVNHashRing.updated_at >= limit)

    if from_host:
        query = query.filter_by(hostname=CONF.host)

    return query


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_READER
def get_active_nodes(context, interval, group_name, from_host=False):
    query = _get_nodes_query(context, interval, group_name,
                             from_host=from_host)
    return query.all()


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_READER
def count_offline_nodes(context, interval, group_name):
    query = _get_nodes_query(context, interval, group_name, offline=True)
    return query.count()


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_READER
def count_nodes_from_host(context, group_name):
    query = context.session.query(ovn_models.OVNHashRing).filter(
        ovn_models.OVNHashRing.group_name == group_name,
        ovn_models.OVNHashRing.hostname == CONF.host)
    return query.count()



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import os
import socketserver
import threading
import time

from neutron_lib import constants
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import fileutils
from oslo_utils import netutils
import webob

from neutron.agent.linux import utils as agent_utils
from neutron.notifiers import batch_notifier

LOG = logging.getLogger(__name__)

KEEPALIVED_STATE_CHANGE_SERVER_BACKLOG = 4096

TRANSLATION_MAP = {'primary': constants.HA_ROUTER_STATE_ACTIVE,
                   'backup': constants.HA_ROUTER_STATE_STANDBY,
                   'fault': constants.HA_ROUTER_STATE_STANDBY,
                   'unknown': constants.HA_ROUTER_STATE_UNKNOWN}

REPLY = """HTTP/1.1 200 OK
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Location': http://127.0.0.1/"""


class KeepalivedStateChangeHandler(socketserver.StreamRequestHandler):
    _agent = None

    def handle(self):
        try:
            request = self.request.recv(4096)
            f_request = io.BytesIO(request)
            req = webob.Request.from_file(f_request)
            router_id = req.headers.get('X-Neutron-Router-Id')
            state = req.headers.get('X-Neutron-State')
            self.enqueue(router_id, state)
            reply = encodeutils.to_utf8(REPLY)
            self.wfile.write(reply)
        except Exception as exc:
            LOG.exception('Error while receiving data.')
            raise exc

    def enqueue(self, router_id, state):
        LOG.debug('Handling notification for router '
                  '%(router_id)s, state %(state)s', {'router_id': router_id,
                                                     'state': state})
        self._agent.enqueue_state_change(router_id, state)


class L3AgentKeepalivedStateChangeServer:
    def __init__(self, agent, conf):
        self.agent = agent
        self.conf = conf
        self._server = None
        agent_utils.ensure_directory_exists_without_file(
            self.get_keepalived_state_change_socket_path(self.conf))

    @classmethod
    def get_keepalived_state_change_socket_path(cls, conf):
        return os.path.join(conf.state_path, 'keepalived-state-change')

    def run(self):
        KeepalivedStateChangeHandler._agent = self.agent
        self._server = agent_utils.UnixDomainWSGIThreadServer(
            'neutron-keepalived-state-change',
            KeepalivedStateChangeHandler,
            self.get_keepalived_state_change_socket_path(self.conf),
        )
        self._server.run()

    def wait(self):
        self._server.wait()


class AgentMixin:
    def __init__(self, host):
        self._init_ha_conf_path()
        super().__init__(host)

    def init_host(self):
        super().init_host()
        # BatchNotifier queue is needed to ensure that the HA router
        # state change sequence is under the proper order.
        self.state_change_notifier = batch_notifier.BatchNotifier(
            self._calculate_batch_duration(), self.notify_server)
        notifications_server = threading.Thread(
            target=self._start_keepalived_notifications_server)
        notifications_server.start()
        self._transition_states = {}
        self._transition_state_mutex = threading.Lock()

    def _get_router_info(self, router_id):
        try:
            return self.router_info[router_id]
        except KeyError:
            LOG.info('Router %s is not managed by this agent. It was '
                     'possibly deleted concurrently.', router_id)

    def check_ha_state_for_router(self, router_id, current_state):
        ri = self._get_router_info(router_id)
        if not ri:
            return
        ha_state = ri.ha_state
        if current_state != TRANSLATION_MAP[ha_state]:
            LOG.debug("Updating server with state %(state)s for router "
                      "%(router_id)s", {'router_id': router_id,
                                        'state': ha_state})
            self.state_change_notifier.queue_event((router_id, ha_state))

    def _start_keepalived_notifications_server(self):
        state_change_server = (
            L3AgentKeepalivedStateChangeServer(self, self.conf))
        state_change_server.run()
        state_change_server.wait()

    def _calculate_batch_duration(self):
        # Set the BatchNotifier interval to ha_vrrp_advert_int,
        # default 2 seconds.
        return self.conf.ha_vrrp_advert_int

    def _update_transition_state(self, router_id, new_state=None):
        with self._transition_state_mutex:
            transition_state = self._transition_states.get(router_id)
            if new_state:
                self._transition_states[router_id] = new_state
            else:
                self._transition_states.pop(router_id, None)
        return transition_state

    def enqueue_state_change(self, router_id, state):
        """Inform the server about the new router state

        This function will also update the metadata proxy, the radvd daemon and
        inform to the L3 extensions. If the HA router changes to "primary",
        this transition will be delayed for at least "ha_vrrp_advert_int"
        seconds. When the "primary" router transitions to "backup",
        "keepalived" will set the rest of HA routers to "primary" until it
        decides which one should be the only "primary". The transition from
        "backup" to "primary" and then to "backup" again, should not be
        registered in the Neutron server.

        :param router_id: router ID
        :param state: ['primary', 'backup']
        """
        if not self._update_transition_state(router_id, state):
            LOG.debug("Enqueueing router's %s state change to %s",
                      router_id, state)
            state_change = threading.Thread(target=self._enqueue_state_change,
                                            args=(router_id, state))
            state_change.start()
            # TODO(ralonsoh): remove once the eventlet deprecation is finished.
            time.sleep(0)

    def _enqueue_state_change(self, router_id, state):
        # NOTE(ralonsoh): move 'primary' and 'backup' constants to n-lib
        if state == 'primary':
            time.sleep(self.conf.ha_vrrp_advert_int)
        transition_state = self._update_transition_state(router_id)
        if transition_state != state:
            # If the current "transition state" is not the initial "state" sent
            # to update the router, that means the actual router state is the
            # same as the "transition state" (e.g.: backup-->primary-->backup).
            LOG.debug("Current transition state of router %s: %s; "
                      "Initial state was: %s",
                      router_id, transition_state, state)
            return

        ri = self._get_router_info(router_id)
        if ri is None:
            return

        state_change_data = {
            "router_id": router_id, "state": state, "host": ri.agent.host,
            "enable_ndp_proxy": ri.router.get("enable_ndp_proxy", False)}
        LOG.info('Router %(router_id)s transitioned to %(state)s on '
                 'agent %(host)s; NDP proxy enabled: %(enable_ndp_proxy)s',
                 state_change_data)

        # TODO(dalvarez): Fix bug 1677279 by moving the IPv6 parameters
        # configuration to keepalived-state-change in order to remove the
        # dependency that currently exists on l3-agent running for the IPv6
        # failover.
        ri.ha_state = state
        self._configure_ipv6_params(ri, state)
        if self.conf.enable_metadata_proxy:
            self._update_metadata_proxy(ri, router_id, state)
        self._update_radvd_daemon(ri, state)
        self.state_change_notifier.queue_event((router_id, state))
        self.l3_ext_manager.ha_state_change(self.context, state_change_data)

    def _configure_ipv6_params(self, ri, state):
        if not self.use_ipv6:
            return

        ipv6_forwarding_enable = state == 'primary'
        if ri.router.get('distributed', False):
            namespace = ri.ha_namespace
        else:
            namespace = ri.ns_name

        if ipv6_forwarding_enable:
            ri.driver.configure_ipv6_forwarding(
                namespace, 'all', ipv6_forwarding_enable)

        # If ipv6 is enabled on the platform, ipv6_gateway config flag is
        # not set and external_network associated to the router does not
        # include any IPv6 subnet, enable the gateway interface to accept
        # Router Advts from upstream router for default route on primary
        # instances as well as ipv6 forwarding. Otherwise, disable them.
        ex_gw_port_id = ri.ex_gw_port and ri.ex_gw_port['id']
        if ex_gw_port_id:
            interface_name = ri.get_external_device_name(ex_gw_port_id)
            ri._configure_ipv6_params_on_gw(
                ri.ex_gw_port, namespace, interface_name,
                ipv6_forwarding_enable)

    def _update_metadata_proxy(self, ri, router_id, state):
        # NOTE(slaweq): Since the metadata proxy is spawned in the qrouter
        # namespace and not in the snat namespace, even standby DVR-HA
        # routers needs to serve metadata requests to local ports.
        if state == 'primary' or ri.router.get('distributed', False):
            LOG.debug('Spawning metadata proxy for router %s', router_id)
            spawn_kwargs = {}
            if netutils.is_ipv6_enabled():
                spawn_kwargs['bind_address'] = '::'
            self.metadata_driver.spawn_monitored_metadata_proxy(
                self.process_monitor, ri.ns_name, self.conf.metadata_port,
                self.conf, router_id=ri.router_id, **spawn_kwargs)
        else:
            LOG.debug('Closing metadata proxy for router %s', router_id)
            self.metadata_driver.destroy_monitored_metadata_proxy(
                self.process_monitor, ri.router_id, self.conf, ri.ns_name)

    def _update_radvd_daemon(self, ri, state):
        # Radvd has to be spawned only on the primary HA Router. If there are
        # any state transitions, we enable/disable radvd accordingly.
        if state == 'primary':
            ri.enable_radvd()
        else:
            ri.disable_radvd()

    def notify_server(self, batched_events):
        translated_states = {router_id: TRANSLATION_MAP[state] for
                             router_id, state in batched_events}
        LOG.debug('Updating server with HA routers states %s',
                  translated_states)
        self.plugin_rpc.update_ha_routers_states(
            self.context, translated_states)

    def _init_ha_conf_path(self):
        ha_full_path = os.path.dirname("/%s/" % self.conf.ha_confs_path)
        fileutils.ensure_tree(ha_full_path, mode=0o755)



# Copyright (c) 2015 UnitedStack, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from textwrap import dedent
from unittest import mock

from neutron.agent.linux import conntrackd
from neutron.tests.unit.agent.l3.test_agent import \
    BasicRouterOperationsFramework
from neutron.tests.unit.agent.l3.test_agent import FAKE_ID


class ConntrackdConfigTestCase(BasicRouterOperationsFramework):

    @staticmethod
    def get_expected(ha_confs_path):
        return dedent(
            """
            General {{
                HashSize 32768
                HashLimit 131072
                Syslog on
                LockFile {conf_path}/{uuid}/conntrackd.lock
                UNIX {{
                    Path {conf_path}/{uuid}/conntrackd.ctl
                    Backlog 20
                }}
                SocketBufferSize 262142
                SocketBufferSizeMaxGrown 655355
                Filter From Kernelspace {{
                    Protocol Accept {{
                        TCP
                        SCTP
                        DCCP
                        UDP
                        ICMP
                        IPv6-ICMP
                    }}
                    Address Ignore {{
                        IPv4_address 127.0.0.1
                        IPv6_address ::1
                        IPv4_address 192.168.0.5
                    }}
                }}
            }}
            Sync {{
                Mode FTFW {{
                }}
                Multicast Default {{
                    IPv4_address 225.0.0.50
                    IPv4_interface 192.168.0.5
                    Group 3783
                    Interface eth0
                    SndSocketBuffer 24985600
                    RcvSocketBuffer 24985600
                    Checksum on
                }}
            }}""".format(conf_path=ha_confs_path, uuid=FAKE_ID))

    def get_manager(self):
        return conntrackd.ConntrackdManager(
            FAKE_ID,
            self.process_monitor,
            self.conf,
            '192.168.0.5',
            3,
            'eth0',
        )

    def test_build_config(self):
        conntrackd = self.get_manager()

        with mock.patch('os.makedirs'):
            config = conntrackd.build_config()
            self.assertMultiLineEqual(
                ConntrackdConfigTestCase.get_expected(self.conf.ha_confs_path),
                config
            )

    def test_max_file_path_len(self):
        """The shortest file path affected by this will be the LockFile path.
        There "/<uuid>/conntrackd.lock" is appended and should in total not
        exceed 255 characters. So the maximum length for ha_confs_path is 202
        characters.
        """

        with mock.patch('os.makedirs'):
            self.conf.set_override('ha_confs_path', '/' + 'a' * 202)
            conntrackd = self.get_manager()
            self.assertRaisesRegex(
                ValueError,
                'maximum length of 255 characters.',
                conntrackd.build_config,
            )

            # If the path is below the file path limit, the UNIX socket path
            # limit is hit.
            self.conf.set_override('ha_confs_path', '/' + 'a' * 201)
            conntrackd = self.get_manager()
            self.assertRaisesRegex(
                ValueError,
                'maximum length of 107 characters.',
                conntrackd.build_config,
            )

    def test_max_socket_path_len(self):
        """The UNIX socket path has a shorter maximum length of 107
        characters. With "/<uuid>/conntrackd.ctl" appended this means the
        maximum length for ha_confs_path is 55 characters.
        """

        with mock.patch('os.makedirs'):
            self.conf.set_override('ha_confs_path', '/' + 'a' * 55)
            conntrackd = self.get_manager()
            self.assertRaisesRegex(
                ValueError,
                'maximum length of 107 characters.',
                conntrackd.build_config,
            )

            self.conf.set_override('ha_confs_path', '/' + 'a' * 54)
            conntrackd = self.get_manager()
            conntrackd.build_config()



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

import netaddr
from neutron_lib import constants
from oslo_log import log

from neutron.agent.common import ovs_lib
from neutron.agent.linux import ip_lib
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.services.bgp import constants as bgp_const

LOG = log.getLogger(__name__)


class Bridge:
    def __init__(self, bgp_agent_api, name):
        self.bgp_agent_api = bgp_agent_api
        self.name = name
        self.ovs_bridge = ovs_lib.OVSBridge(name)

    @property
    def ovs_idl(self):
        return self.bgp_agent_api.agent_api.ovs_idl

    @property
    def sb_idl(self):
        return self.bgp_agent_api.agent_api.sb_idl

    def _apply_flows_as_bundle(self, flows):
        """Apply multiple OpenFlow rules as a bundle using temporary file"""
        with tempfile.NamedTemporaryFile(mode='w', suffix='.flows',
                                        prefix='bgp_', delete=True) as f:
            f.write("delete\n")
            for flow in flows:
                f.write(f"{flow}\n")
            f.flush()

            LOG.debug("Applying %d BGP flows as bundle to bridge %s",
                      len(flows), self.name)

            self.ovs_bridge.run_ofctl(
                "add-flows", ["--bundle", f.name])


class BGPChassisBridge(Bridge):
    """BGP Bridge

    The BGP bridge is the provider bridge that connects a chassis to a BGP
    physical interface connected to a BGP peer, typically a leaf switch.
    """
    def __init__(self, bgp_agent_api, name):
        super().__init__(bgp_agent_api, name)
        self._requirements = [
            ('bridge does not exist', 'exists'),
            ('NIC ofport is not set', 'nic_ofport'),
            ('patch port ofport is not set', 'patch_port_ofport'),
            ('LRP MAC is not set', 'lrp_mac'),
        ]

    def __str__(self):
        return f"BGPChassisBridge(name={self.name})"

    __repr__ = __str__

    def check_requirements_for_flows_met(self):
        for msg, requirement in self._requirements:
            if not getattr(self, requirement):
                LOG.debug(
                    "Bridge %s: %s, skipping installing flows",
                    self.name, msg)
                return False
        return True

    @property
    def patch_port_ofport(self):
        patch_ports_ofports = self.ovs_bridge.get_iface_ofports_by_type(
            'patch')
        if len(patch_ports_ofports) > 1:
            LOG.warning("The patch port for bridge %s has multiple ofports: "
                        "%s, using the first one",
                        self.name, patch_ports_ofports)
        try:
            return patch_ports_ofports[0]
        except IndexError:
            LOG.debug("The patch port for bridge %s does not exist yet",
                      self.name)
            return None

    @property
    def nic_ofport(self):
        # REVISIT(jlibosva): we can consider supporting OVS bonds too
        nics_ofports = self.ovs_bridge.get_iface_ofports_by_type('')
        if len(nics_ofports) != 1:
            LOG.warning("Expected 1 NIC for bridge %s, got %s",
                        self.name, len(nics_ofports))
            return None
        return nics_ofports[0]

    @property
    def lrp_mac(self):
        ext_ids = {bgp_const.LRP_NETWORK_NAME_EXT_ID_KEY: self.name}
        port_bindings = self.sb_idl.db_find_rows(
            'Port_Binding',
            ('type', '=', ovn_const.PB_TYPE_L3GATEWAY),
            ('external_ids', '=', ext_ids)).execute(
                check_error=True)
        if port_bindings:
            pb = port_bindings[0]
            try:
                return ovn_utils.get_mac_and_ips_from_port_binding(pb)[0]
            except ValueError:
                LOG.error("Failed to get MAC address from port binding %s",
                          pb.logical_port)

        LOG.debug("LRP MAC does not exist yet for %s", self.name)

    @property
    def exists(self):
        return self.ovs_bridge.bridge_exists(self.name)

    @property
    def ips(self):
        return [netaddr.IPNetwork(ip['cidr'])
                for ip in ip_lib.get_devices_with_ip(
                    namespace=None, name=self.name)]

    def _get_flows_for_icmpv6(self):
        """Ingress flows for ICMPv6.

        We don't know if ND or RA related packets are from the host or the
        per chassis BGP router, so we flood the traffic to both.
        """
        flows = [(f"priority=100,in_port={self.nic_ofport},icmp6,"
                 f"icmp_type={icmp_type} "
                 f"actions=NORMAL,mod_dl_dst:{self.lrp_mac},"
                 f"output:{self.patch_port_ofport}")
                 for icmp_type in [133, 134, 135, 136]]

        return flows

    def _get_flows_for_host_ips(self):
        LOG.debug("Adding flows to direct traffic to the host from the NIC "
                  "port %s on bridge %s", self.nic_ofport, self.name)
        # Allow IPv6 link-local traffic
        flows = [f"priority=100,ipv6,in_port={self.nic_ofport},"
                 f"ipv6_dst=fe80::/64 actions=NORMAL,"
                 f"mod_dl_dst:{self.lrp_mac},output:{self.patch_port_ofport}"]

        # Direct traffic meant for the host IPs
        for host_ip in self.bgp_agent_api.hostdev_ips + self.ips:
            if host_ip.version == constants.IP_VERSION_4:
                flows.append(f"priority=100,ip,in_port={self.nic_ofport},"
                             f"nw_dst={host_ip.ip} actions=NORMAL")
            elif host_ip.version == constants.IP_VERSION_6:
                flows.append(f"priority=100,ipv6,in_port={self.nic_ofport},"
                             f"ipv6_dst={host_ip.ip} actions=NORMAL")

        return flows

    def configure_flows(self):
        # The resulting openflows rules that will be written to a temporary
        # file and applied to the bridge.
        if not self.check_requirements_for_flows_met():
            LOG.error("Some of the requirements to install flows on bridge "
                      "%s are missing, skipping", self.name)
            return

        LOG.debug("Configuring BGP bridge flows for %s", self.name)
        # Allow ARP and ICMPv6
        flows = [
            "priority=100,arp actions=NORMAL",

            # Put the destination MAC of the LRP on the per-chassis router for
            # any remaining traffic going from the NIC.
            (f"priority=80,in_port={self.nic_ofport},"
             f"actions=mod_dl_dst:{self.lrp_mac},"
             f"output:{self.patch_port_ofport}"),

            # Any traffic coming from the patch port should go out.
            (f"priority=100,in_port={self.patch_port_ofport},"
             f"actions=NORMAL"),

            # Allow all other traffic
            "priority=0, actions=NORMAL",
        ]

        flows.extend(self._get_flows_for_host_ips())
        flows.extend(self._get_flows_for_icmpv6())

        try:
            self._apply_flows_as_bundle(flows)
        except Exception as e:
            LOG.error("Failed to configure BGP flows on bridge %s: %s: %s",
                      self.name, e, flows)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = """
The QoS API now supports project scope and default roles.
"""
RESOURCE_PATH = '/qos/policies/{id}'
TAGS_PATH = RESOURCE_PATH + '/tags'
TAG_PATH = RESOURCE_PATH + '/tags/{tag_id}'

ACTION_GET_TAGS = [
    {'method': 'GET', 'path': TAGS_PATH},
    {'method': 'GET', 'path': TAG_PATH},
]
ACTION_PUT_TAGS = [
    {'method': 'PUT', 'path': TAGS_PATH},
    {'method': 'PUT', 'path': TAG_PATH},
]
ACTION_POST_TAGS = [
    {'method': 'POST', 'path': TAGS_PATH},
]
ACTION_DELETE_TAGS = [
    {'method': 'DELETE', 'path': TAGS_PATH},
    {'method': 'DELETE', 'path': TAG_PATH},
]


rules = [
    policy.RuleDefault(
        'shared_qos_policy',
        'field:policies:shared=True',
        description='Rule of shared qos policy'),
    policy.DocumentedRuleDefault(
        name='get_policy',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            'rule:shared_qos_policy'
        ),
        scope_types=['project'],
        description='Get QoS policies',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies',
            },
            {
                'method': 'GET',
                'path': '/qos/policies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_policy',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_policy:tags',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            'rule:shared_qos_policy'
        ),
        scope_types=['project'],
        description='Get QoS policy tags',
        operations=ACTION_GET_TAGS,
        deprecated_rule=policy.DeprecatedRule(
            name='get_policies_tags',
            check_str=neutron_policy.policy_or(
                base.ADMIN_OR_PROJECT_READER,
                'rule:shared_qos_policy'
            ),
            deprecated_reason="Name of the rule is changed.",
            deprecated_since="2025.1")
    ),
    policy.DocumentedRuleDefault(
        name='create_policy',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Create a QoS policy',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_policy',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_policy:tags',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Create the QoS policy tags',
        operations=ACTION_POST_TAGS,
        deprecated_rule=policy.DeprecatedRule(
            name='create_policies_tags',
            check_str=base.ADMIN_OR_PROJECT_MANAGER,
            deprecated_reason="Name of the rule is changed.",
            deprecated_since="2025.1")
    ),
    policy.DocumentedRuleDefault(
        name='update_policy',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Update a QoS policy',
        operations=[
            {
                'method': 'PUT',
                'path': '/qos/policies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_policy',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_policy:tags',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Update the QoS policy tags',
        operations=ACTION_PUT_TAGS,
        deprecated_rule=policy.DeprecatedRule(
            name='update_policies_tags',
            check_str=base.ADMIN_OR_PROJECT_MANAGER,
            deprecated_reason="Name of the rule is changed.",
            deprecated_since="2025.1")
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Delete a QoS policy',
        operations=[
            {
                'method': 'DELETE',
                'path': '/qos/policies/{id}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_policy',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy:tags',
        check_str=base.ADMIN_OR_PROJECT_MANAGER,
        scope_types=['project'],
        description='Delete the QoS policy tags',
        operations=ACTION_DELETE_TAGS,
        deprecated_rule=policy.DeprecatedRule(
            name='delete_policies_tags',
            check_str=base.ADMIN_OR_PROJECT_MANAGER,
            deprecated_reason="Name of the rule is changed.",
            deprecated_since="2025.1")
    ),

    policy.DocumentedRuleDefault(
        name='get_rule_type',
        # NOTE(ralonsoh): it can't be ADMIN_OR_PROJECT_READER constant from the
        # base module because that is using "project_id" in the check string
        # and the rule type resource don't belongs to any project thus such
        # check string would fail enforcement.
        check_str='role:reader',
        scope_types=['project'],
        description='Get available QoS rule types',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/rule-types',
            },
            {
                'method': 'GET',
                'path': '/qos/rule-types/{rule_type}',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_rule_type',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='get_policy_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS bandwidth limit rule',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies/{policy_id}/bandwidth_limit_rules',
            },
            {
                'method': 'GET',
                'path': ('/qos/policies/{policy_id}/'
                         'bandwidth_limit_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_policy_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_policy_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Create a QoS bandwidth limit rule',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies/{policy_id}/bandwidth_limit_rules',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_policy_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_policy_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS bandwidth limit rule',
        operations=[
            {
                'method': 'PUT',
                'path': ('/qos/policies/{policy_id}/'
                         'bandwidth_limit_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_policy_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS bandwidth limit rule',
        operations=[
            {
                'method': 'DELETE',
                'path': ('/qos/policies/{policy_id}/'
                         'bandwidth_limit_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_policy_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='get_policy_packet_rate_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS packet rate limit rule',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies/{policy_id}/packet_rate_limit_rules',
            },
            {
                'method': 'GET',
                'path': ('/qos/policies/{policy_id}/'
                         'packet_rate_limit_rules/{rule_id}'),
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        name='create_policy_packet_rate_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Create a QoS packet rate limit rule',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies/{policy_id}/packet_rate_limit_rules',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        name='update_policy_packet_rate_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS packet rate limit rule',
        operations=[
            {
                'method': 'PUT',
                'path': ('/qos/policies/{policy_id}/'
                         'packet_rate_limit_rules/{rule_id}'),
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy_packet_rate_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS packet rate limit rule',
        operations=[
            {
                'method': 'DELETE',
                'path': ('/qos/policies/{policy_id}/'
                         'packet_rate_limit_rules/{rule_id}'),
            },
        ]
    ),

    policy.DocumentedRuleDefault(
        name='get_policy_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS DSCP marking rule',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies/{policy_id}/dscp_marking_rules',
            },
            {
                'method': 'GET',
                'path': ('/qos/policies/{policy_id}/'
                         'dscp_marking_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_policy_dscp_marking_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_policy_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Create a QoS DSCP marking rule',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies/{policy_id}/dscp_marking_rules',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_policy_dscp_marking_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_policy_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS DSCP marking rule',
        operations=[
            {
                'method': 'PUT',
                'path': ('/qos/policies/{policy_id}/'
                         'dscp_marking_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_policy_dscp_marking_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS DSCP marking rule',
        operations=[
            {
                'method': 'DELETE',
                'path': ('/qos/policies/{policy_id}/'
                         'dscp_marking_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_policy_dscp_marking_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),

    policy.DocumentedRuleDefault(
        name='get_policy_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS minimum bandwidth rule',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies/{policy_id}/minimum_bandwidth_rules',
            },
            {
                'method': 'GET',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_bandwidth_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_policy_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_policy_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Create a QoS minimum bandwidth rule',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies/{policy_id}/minimum_bandwidth_rules',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_policy_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_policy_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS minimum bandwidth rule',
        operations=[
            {
                'method': 'PUT',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_bandwidth_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_policy_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS minimum bandwidth rule',
        operations=[
            {
                'method': 'DELETE',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_bandwidth_rules/{rule_id}'),
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_policy_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_policy_minimum_packet_rate_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS minimum packet rate rule',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/policies/{policy_id}/minimum_packet_rate_rules',
            },
            {
                'method': 'GET',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_packet_rate_rules/{rule_id}'),
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='create_policy_minimum_packet_rate_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Create a QoS minimum packet rate rule',
        operations=[
            {
                'method': 'POST',
                'path': '/qos/policies/{policy_id}/minimum_packet_rate_rules',
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='update_policy_minimum_packet_rate_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS minimum packet rate rule',
        operations=[
            {
                'method': 'PUT',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_packet_rate_rules/{rule_id}'),
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='delete_policy_minimum_packet_rate_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS minimum packet rate rule',
        operations=[
            {
                'method': 'DELETE',
                'path': ('/qos/policies/{policy_id}/'
                         'minimum_packet_rate_rules/{rule_id}'),
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name='get_alias_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS bandwidth limit rule through alias',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/alias_bandwidth_limit_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_alias_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_alias_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS bandwidth limit rule through alias',
        operations=[
            {
                'method': 'PUT',
                'path': '/qos/alias_bandwidth_limit_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_alias_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_alias_bandwidth_limit_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS bandwidth limit rule through alias',
        operations=[
            {
                'method': 'DELETE',
                'path': '/qos/alias_bandwidth_limit_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_alias_bandwidth_limit_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_alias_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS DSCP marking rule through alias',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/alias_dscp_marking_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_alias_dscp_marking_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_alias_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS DSCP marking rule through alias',
        operations=[
            {
                'method': 'PUT',
                'path': '/qos/alias_dscp_marking_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_alias_dscp_marking_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_alias_dscp_marking_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS DSCP marking rule through alias',
        operations=[
            {
                'method': 'DELETE',
                'path': '/qos/alias_dscp_marking_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_alias_dscp_marking_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_alias_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_READER,
        scope_types=['project'],
        description='Get a QoS minimum bandwidth rule through alias',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/alias_minimum_bandwidth_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_alias_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ANY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='update_alias_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Update a QoS minimum bandwidth rule through alias',
        operations=[
            {
                'method': 'PUT',
                'path': '/qos/alias_minimum_bandwidth_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='update_alias_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_alias_minimum_bandwidth_rule',
        check_str=base.ADMIN_OR_PARENT_OWNER_MANAGER,
        scope_types=['project'],
        description='Delete a QoS minimum bandwidth rule through alias',
        operations=[
            {
                'method': 'DELETE',
                'path': '/qos/alias_minimum_bandwidth_rules/{rule_id}/',
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_alias_minimum_bandwidth_rule',
            check_str=neutron_policy.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_alias_minimum_packet_rate_rule',
        check_str='rule:get_policy_minimum_packet_rate_rule',
        scope_types=['project'],
        description='Get a QoS minimum packet rate rule through alias',
        operations=[
            {
                'method': 'GET',
                'path': '/qos/alias_minimum_packet_rate_rules/{rule_id}/',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        name='update_alias_minimum_packet_rate_rule',
        check_str='rule:update_policy_minimum_packet_rate_rule',
        scope_types=['project'],
        description='Update a QoS minimum packet rate rule through alias',
        operations=[
            {
                'method': 'PUT',
                'path': '/qos/alias_minimum_packet_rate_rules/{rule_id}/',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        name='delete_alias_minimum_packet_rate_rule',
        check_str='rule:delete_policy_minimum_packet_rate_rule',
        scope_types=['project'],
        description='Delete a QoS minimum packet rate rule through alias',
        operations=[
            {
                'method': 'DELETE',
                'path': '/qos/alias_minimum_packet_rate_rules/{rule_id}/',
            },
        ]
    ),
]


def list_rules():
    return rules



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_log import log

from neutron.agent.linux import ip_lib
from neutron.agent.ovn.agent import ovsdb
from neutron.agent.ovn.extensions.bgp import bridge
from neutron.agent.ovn.extensions.bgp import commands
from neutron.agent.ovn.extensions.bgp import events
from neutron.agent.ovn.extensions import extension_manager as ovn_ext_mgr

LOG = log.getLogger(__name__)


class BGPAgentExtension(ovn_ext_mgr.OVNAgentExtension):
    def __init__(self):
        super().__init__()
        # A map of bridge names to the bridge object
        # Example: {
        #     'br-eth1': BGPChassisBridge('br-eth1'),
        #     'br-eth2': BGPChassisBridge('br-eth2'),
        # }
        self.bgp_bridges = {}
        self.hostdev_name = ip_lib.LOOPBACK_DEVNAME

    @property
    def name(self):
        return "BGP agent extension"

    @property
    def ovs_idl_events(self):
        return [
            events.CreateLocalOVSEvent,
            events.UpdateLocalOVSEvent,
            events.NewBgpBridgeEvent,
        ]

    @property
    def nb_idl_tables(self):
        return []

    @property
    def nb_idl_events(self):
        return []

    @property
    def sb_idl_tables(self):
        return [
            'Port_Binding',
        ]

    @property
    def sb_idl_events(self):
        return [
            events.PortBindingLrpMacEvent,
        ]

    @property
    def chassis_name(self):
        return ovsdb.get_own_chassis_name(self.agent_api.ovs_idl)

    @property
    def hostdev_ips(self):
        """Return the IP addresses configured on a host device.

        The host device is typically a loopback device and all host IPs should
        be configured there.
        """
        return [cidr for dev in ip_lib.get_devices_with_ip(
                    namespace=None, name=self.hostdev_name)
                if not (cidr := netaddr.IPNetwork(dev['cidr'])).is_loopback()]

    def create_bgp_bridge(self, bridge_name):
        bgp_bridge = bridge.BGPChassisBridge(self, bridge_name)
        self.bgp_bridges[bridge_name] = bgp_bridge
        return bgp_bridge

    def set_chassis_bgp_bridges(self, bridge_name_list):
        commands.SetChassisBgpBridgesCommand(
            self.agent_api.sb_idl,
            self.chassis_name,
            bridge_name_list
        ).execute(check_error=True)

    def watch_port_created_event(self, bgp_bridge, port_type):
        # Check the port doesn't exist on the bridge
        ports_ofports = bgp_bridge.ovs_bridge.get_iface_ofports_by_type(
            port_type)

        if not ports_ofports:
            LOG.debug("Waiting for a %s port creation on bridge %s",
                      port_type, bgp_bridge.name)
            event_handler = self.agent_api.ovs_idl.idl.notify_handler
            event = events.BGPBridgePortCreatedEvent(
                self.agent_api, bgp_bridge.name, port_type)
            event_handler.watch_event(event)

            # Check the port again in case it was created in the meantime
            ports_ofports = (
                bgp_bridge.ovs_bridge.get_iface_ofports_by_type(port_type))

            # FIXME(jlibosva): Check if there could be a race condition here
            #                  where we receive the event, it configures the
            #                  flows but then we still check here and configure
            #                  the flows again.
            if ports_ofports:
                LOG.debug(
                    "The %s port was created in the meantime on bridge %s "
                    "with ofport %d, removing the onetime event from the "
                    "queue.", port_type, bgp_bridge.name, ports_ofports[0])
                event_handler.unwatch_event(event)
                if bgp_bridge.check_requirements_for_flows_met():
                    bgp_bridge.configure_flows()
        else:
            LOG.debug("The BGP bridge %s already has a %s port with ofport"
                      " %d", bgp_bridge.name, port_type, ports_ofports[0])
            if bgp_bridge.check_requirements_for_flows_met():
                bgp_bridge.configure_flows()



# Copyright (c) 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port_trusted_vif
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const

from neutron.objects.port.extensions import port_trusted as trusted_obj


class PortTrustedDbMixin:
    """Mixin class to add trusted extension to a port"""

    @staticmethod
    def _set_portbinding_profile(data, trusted):
        try:
            data[portbindings.PROFILE]['trusted'] = trusted
        except (AttributeError, KeyError):
            data[portbindings.PROFILE] = {'trusted': trusted}

    def _process_create_port(self, context, data, result):
        trusted = data.get(port_trusted_vif.TRUSTED_VIF)
        if trusted is n_const.ATTR_NOT_SPECIFIED:
            result[port_trusted_vif.TRUSTED_VIF] = None
            return

        obj = trusted_obj.PortTrusted(
            context, port_id=result['id'], trusted=trusted)
        obj.create()
        result[port_trusted_vif.TRUSTED_VIF] = trusted
        self._set_portbinding_profile(result, trusted)

    def _process_update_port(self, context, data, result):
        trusted = data.get(port_trusted_vif.TRUSTED_VIF)
        if trusted is None or trusted is n_const.ATTR_NOT_SPECIFIED:
            result[port_trusted_vif.TRUSTED_VIF] = None
            return

        obj = trusted_obj.PortTrusted.get_object(
            context, port_id=result['id'])
        if obj:
            obj.trusted = trusted
            obj.update()
            result[port_trusted_vif.TRUSTED_VIF] = trusted
            self._set_portbinding_profile(result, trusted)
        else:
            self._process_create_port(context, data, result)

    def _extend_port_dict(self, response_data, db_data):
        if db_data.trusted is not None:
            trusted = db_data.trusted.trusted
            response_data[port_trusted_vif.TRUSTED_VIF] = trusted
            self._set_portbinding_profile(response_data, trusted)
        else:
            response_data[port_trusted_vif.TRUSTED_VIF] = None



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum

from oslo_log import log

from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.services.bgp import commands
from neutron.services.bgp import events
from neutron.services.bgp import helpers
from neutron.services.bgp import ovn


LOG = log.getLogger(__name__)


class BGPTopologyReconciler:
    class BGPReconcilerResource(enum.Enum):
        CHASSIS_BGP_BRIDGES = 'chassis-bgp-bridges'

        def __str__(self):
            return self.value

    def __init__(self):
        self.nb_api = ovn.OvnNbIdl(
            ovn_conf.get_ovn_nb_connection(),
            self.nb_events).start(
                timeout=ovn_conf.get_ovn_ovsdb_timeout())
        self.sb_api = ovn.OvnSbIdl(
            ovn_conf.get_ovn_sb_connection(),
            self.sb_events).start(
                timeout=ovn_conf.get_ovn_ovsdb_timeout())

    def stop(self):
        self.nb_api.stop()
        self.sb_api.stop()

    @property
    def resource_map(self):
        return {
            self.BGPReconcilerResource.CHASSIS_BGP_BRIDGES:
                self.reconcile_chassis_bgp_bridges,
        }

    @property
    def nb_events(self):
        return [
        ]

    @property
    def sb_events(self):
        return [
            events.BGPChassisBridgesUpdateEvent(self),
        ]

    def full_sync(self):
        if not self.nb_api.ovsdb_connection.idl.is_lock_contended:
            LOG.info("Full BGP topology synchronization started")
            # First make sure all chassis are indexed
            commands.FullSyncBGPTopologyCommand(
                self.nb_api, self.sb_api).execute(check_error=True)
            LOG.info(
                "Full BGP topology synchronization completed successfully")
        else:
            LOG.info("Full BGP topology synchronization already in progress")

    def reconcile(self, resource, trigger):
        try:
            self.resource_map[resource](trigger)
        except KeyError:
            LOG.error("Resource %s not found in reconciler resource map",
                      resource)

    def reconcile_chassis_bgp_bridges(self, chassis):
        for bgp_bridge in helpers.get_chassis_bgp_bridges(chassis):
            commands.ReconcileChassisPeerCommand(
                self.nb_api,
                chassis,
                network_name=bgp_bridge,
            ).execute(check_error=True)
